[Unit]
Description=CFEngine Enterprise event reaction daemon
PartOf=cfengine3.service
ConditionPathExists={{{vars.sys.bindir}}}/cf-reactor
ConditionPathExists={{{vars.sys.default_policy_path}}}
After=syslog.target
After=network.target

Wants=cf-postgres.service
After=cf-postgres.service

[Service]
Type=simple
ExecStart={{{vars.sys.bindir}}}/cf-reactor --no-fork
Restart=always
RestartSec=10

[Install]
WantedBy=multi-user.target
WantedBy=cfengine3.service
